/* cosh.c */
/* Copyright (C) Advanced RISC Machines Limited, 1994. All rights reserved. */

/*
 * RCS $Revision: 1.5 $
 * Checkin $Date: 1995/06/01 16:39:57 $
 * Revising $Author: mwilliam $
 */

#include "mathlib.h"

static ip_number sinh_lnv =  { 0x00003ffe, 0xb1730000, 0x00000000 };
static ip_number sinh_vm2 =  { 0x00003ffc, 0xfffe2ff1, 0x483b9d27 };
static ip_number sinh_v2m1 = { 0x00003fee, 0xe8089758, 0x1016b37d };

static ip_number one =       { 0x00003fff, 0x80000000, 0x00000000 };
static ip_number half =      { 0x00003ffe, 0x80000000, 0x00000000 };

double cosh(double x)
{
  ip_number ix=_d2e(x);

  if (fp_uncommon(ix)) goto uncommon;
  fp_abs(ix);

  if (fp_grpow(ix,0)) {
    ix=_esub(ix,&sinh_lnv);
    ix=_exp(ix);
    if (fp_error(ix)) goto error;
    if (!fp_grpow(ix,32)) {
      ix=_eadd(_erdv(ix,&sinh_vm2),&ix);
    }
    ix=_eadd(_emul(ix,&sinh_v2m1),&ix);
  } else {
    ix=_exp(ix);
    ix=_emul(_eadd(_erdv(ix,&one),&ix),&half);
  }
  if (!fp_error(ix)) return _e2d(ix);

  return __fp_erange(ix.word.hi);

 uncommon:
  if (!fp_infinity(ix)) return __fp_edom(ix.word.hi);
  /* Drop through */

 error:
  return x;
}
