/* exp.c */
/* Copyright (C) Advanced RISC Machines Limited, 1994. All rights reserved. */

/*
 * RCS $Revision: 1.5 $
 * Checkin $Date: 1995/06/01 16:41:08 $
 * Revising $Author: mwilliam $
 */

#include <stdarg.h>
#include <stdlib.h>

#include <limits.h>
#include <errno.h>
#include <math.h>

#include "mathlib.h"

#ifdef _exp_c

static const ip_number exp_range[] = {
  0x00003fff, 0xb8aa3b29, 0x5c17f0bc, /* 1/ln(2) */
  0x80003ff2, 0xde8082e3, 0x08654362, /* M2 (ln(2)-M1) */
  0x00003ffe, 0xb1800000, 0x00000000  /* M1 */
  };

static const ip_number exp_P[] = {
  0x00003ff1, 0x845a2157, 0x3490f106, /* 2*p2 (p1=0.31555192765684646356e-4) */
  0x00003ff8, 0xf83a5f91, 0x50952c99, /* 2*p1 (p2=0.75753180159422776666e-2) */
  0x00003ffe, 0x80000000, 0x00000000  /* 2*p0 (p0=0.25000000000000000000e+0) */
  };

static const ip_number exp_Q[] = {
  0x00003feb, 0xc99b1867, 0x3490f106, /* 2*q3 (q3=0.75104028399870046114e-6) */
  0x00003ff5, 0xa57862e1, 0x46a6fb39, /* 2*q2 (q2=0.63121894374398503557e-3) */
  0x00003ffb, 0xe8b9428e, 0xfecff592, /* 2*q1 (q1=0.56817302698551221787e-1) */
  0x00003fff, 0x80000000, 0x00000000  /* 2*q0 (q0=0.50000000000000000000e-0) */
  };

static const ip_number half = { 0x3ffe, 0x80000000, 0x0 };
static const ip_number one =  { 0x3fff, 0x80000000, 0x0 };
static const ip_number zero = { 0x0, 0x0, 0x0 };
static const ip_number error_word =  { error_bit, 0x0, 0x0 };

__value_in_regs ip_number _exp(ip_number ix)
{
  range_red g;
  ip_number z,gpz,qz,r;

  if (fp_geqpow(ix,-64)) {
    g=__fp_range_red_by_mod(ix,exp_range);
    if (fp_error(g.x)) goto error;
    z=_esquare(g.x);
    
    gpz=_emul(__fp_poly(z,3,exp_P),&g.x);
    qz=__fp_poly(z,4,exp_Q);
    
    r=_eadd(_erdv(_esub(qz,&gpz),&gpz),&half);
    
    r.word.hi+=g.n+1;
    return r;
  } else {
    return one;
  }

 error:
  if (fp_sign(ix)) return zero; /* Negative */

  errno=ERANGE;
  return error_word;
}

#else
/* declared in mathlib.h */
#endif

#ifdef exp_c

double exp(double x)
{
  ip_number ix;

  ix=_d2e(x);

  if (!fp_uncommon(ix)) {
    ix=_exp(ix);
    if (!fp_error(ix)) return _e2d(ix);
    
    return fp_sign(ix) ? -HUGE_VAL : HUGE_VAL;
  }

  if (fp_infinity(ix)) return __fp_erange(ix.word.hi);
  return __fp_edom(ix.word.hi);
}

#endif
